/*
 * Exec.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("f622148a-93e0-4fbc-9645-e2ead4e5483b")]
    [CommandFlags(CommandFlags.Unsafe | CommandFlags.Standard)]
    [ObjectGroup("nativeEnvironment")]
    internal sealed class Exec : Core
    {
        public Exec(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        OptionDictionary options = new OptionDictionary(
                            new Option[] {
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-debug", null),              // simple switch
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-commandline", null),        // simple switch
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-quoteall", null),           // simple switch
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-ignorestderr", null),       // simple switch
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-keepnewline", null),        // simple switch
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-nocapture", null),          // simple switch
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-shell", null),              // simple switch
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, "-nocarriagereturns", null),  // simple switch
                            new Option(typeof(ExitCode), OptionFlags.MustHaveEnumValue,
                                Index.Invalid, Index.Invalid, "-success", null), // success exit code
                            new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                                Index.Invalid, "-directory", null), // directory name
                            new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                                Index.Invalid, "-exitcode", null),  // varName for exitCode
                            new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                                Index.Invalid, "-stdin", null),     // varName for StdIn input
                            new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                                Index.Invalid, "-stdout", null),    // varName for StdOut output
                            new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                                Index.Invalid, "-stderr", null),    // varName for StdErr output
                            new Option(typeof(EventFlags), OptionFlags.MustHaveEnumValue,
                                Index.Invalid, Index.Invalid, "-eventflags", null),
                            new Option(null, OptionFlags.None, Index.Invalid,
                                Index.Invalid, Option.EndOfOptions, null)
                        });

                        int argumentIndex = Index.Invalid;

                        code = interpreter.GetOptions(options, arguments, 0, 1,
                            Index.Invalid, true, ref argumentIndex, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            if (argumentIndex != Index.Invalid)
                            {
                                bool debug = false;

                                if (options.IsPresent("-debug"))
                                    debug = true;

                                bool commandLine = false;

                                if (options.IsPresent("-commandline"))
                                    commandLine = true;

                                bool quoteAll = false;

                                if (options.IsPresent("-quoteall"))
                                    quoteAll = true;

                                bool captureInput = true;
                                bool captureOutput = true;

                                if (options.IsPresent("-nocapture"))
                                {
                                    captureInput = false;
                                    captureOutput = false;
                                }

                                bool ignoreStdErr = false;

                                if (options.IsPresent("-ignorestderr"))
                                    ignoreStdErr = true;

                                bool keepNewLine = false;

                                if (options.IsPresent("-keepnewline"))
                                    keepNewLine = true;

                                bool carriageReturns = true;

                                if (options.IsPresent("-nocarriagereturns"))
                                    carriageReturns = false;

                                bool useShellExecute = false;

                                if (options.IsPresent("-shell"))
                                    useShellExecute = true;

                                Variant value = null;
                                ExitCode? successExitCode = null;

                                if (options.IsPresent("-success", ref value))
                                    successExitCode = (ExitCode)value.Value;

                                string directory = null;

                                if (options.IsPresent("-directory", ref value))
                                    directory = value.ToString();

                                string exitCodeVarName = null;

                                if (options.IsPresent("-exitcode", ref value))
                                    exitCodeVarName = value.ToString();

                                string stdInVarName = null;

                                if (options.IsPresent("-stdin", ref value))
                                    stdInVarName = value.ToString();

                                string stdOutVarName = null;

                                if (options.IsPresent("-stdout", ref value))
                                    stdOutVarName = value.ToString();

                                string stdErrVarName = null;

                                if (options.IsPresent("-stderr", ref value))
                                    stdErrVarName = value.ToString();

                                EventFlags eventFlags = interpreter.EngineEventFlags;

                                if (options.IsPresent("-eventflags", ref value))
                                    eventFlags = (EventFlags)value.Value;

                                int argumentStopIndex = arguments.Count - 1;
                                bool background = false;

                                if (arguments[arguments.Count - 1] ==
                                        Characters.Ampersand.ToString())
                                {
                                    argumentStopIndex--;
                                    background = true;
                                }

                                string execFileName = arguments[argumentIndex];

                                if (!PathOps.IsRemoteUri(execFileName))
                                    execFileName = PathOps.GetNativePath(execFileName);

                                string execArguments = null;

                                if ((argumentIndex + 1) < arguments.Count)
                                {
                                    if (commandLine)
                                        execArguments = RuntimeOps.BuildCommandLine(
                                            ArgumentList.GetRangeAsStringList(arguments,
                                                argumentIndex + 1, argumentStopIndex),
                                            quoteAll);
                                    else
                                        execArguments = ListOps.Concat(arguments,
                                            argumentIndex + 1, argumentStopIndex);
                                }

                                Result input = null;

                                if ((code == ReturnCode.Ok) && !useShellExecute &&
                                    captureInput && (stdInVarName != null))
                                {
                                    code = interpreter.GetVariableValue(VariableFlags.None,
                                        stdInVarName, ref input, ref result);
                                }

                                if (debug)
                                {
                                    RuntimeOps.DebugTrace(String.Format(
                                        "{0}: Execute: interpreter = {1}, execFileName = {2}, " +
                                        "execArguments = {3}, directory = {4}, input = {5}, " +
                                        "eventFlags = {6}, useShellExecute = {7}, captureInput = {8}, " +
                                        "captureOutput = {9}, ignoreStdErr = {10}, keepNewLine = {11}, " +
                                        "background = {12}{13}", GlobalState.GetCurrentThreadId(),
                                        FormatOps.WrapOrNull(interpreter), FormatOps.WrapOrNull(execFileName),
                                        FormatOps.WrapOrNull(execArguments), FormatOps.WrapOrNull(directory),
                                        FormatOps.WrapOrNull(input), FormatOps.WrapOrNull(eventFlags),
                                        useShellExecute, captureInput, captureOutput, ignoreStdErr,
                                        keepNewLine, background, Environment.NewLine), typeof(Exec).Name);
                                }

                                int processId = 0;
                                ExitCode exitCode = RuntimeOps.SuccessExitCode();
                                Result error = null;

                                if (code == ReturnCode.Ok)
                                {
                                    code = ProcessOps.ExecuteProcess(
                                        interpreter, execFileName, execArguments, directory, input,
                                        eventFlags, useShellExecute, captureOutput, ignoreStdErr,
                                        keepNewLine, background, !background, ref processId,
                                        ref exitCode, ref result, ref error);
                                }

                                if (debug)
                                {
                                    RuntimeOps.DebugTrace(String.Format(
                                        "{0}: Execute: interpreter = {1}, execFileName = {2}, " +
                                        "execArguments = {3}, directory = {4}, input = {5}, " +
                                        "eventFlags = {6}, useShellExecute = {7}, captureInput = {8}, " +
                                        "captureOutput = {9}, ignoreStdErr = {10}, keepNewLine = {11}, " +
                                        "background = {12}, processId = {13}, exitCode = {14}, " +
                                        "result = {15}, error = {16}{17}", GlobalState.GetCurrentThreadId(),
                                        FormatOps.WrapOrNull(interpreter), FormatOps.WrapOrNull(execFileName),
                                        FormatOps.WrapOrNull(execArguments), FormatOps.WrapOrNull(directory),
                                        FormatOps.WrapOrNull(input), FormatOps.WrapOrNull(eventFlags),
                                        useShellExecute, captureInput, captureOutput, ignoreStdErr,
                                        keepNewLine, background, processId, exitCode,
                                        FormatOps.WrapOrNull(true, true, result),
                                        FormatOps.WrapOrNull(true, true, error), Environment.NewLine),
                                        typeof(Exec).Name);
                                }

                                if (code == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Remove all carriage returns from output (leaving
                                    //       only line feeds as line separators)?
                                    //
                                    if (!carriageReturns)
                                    {
                                        if (!String.IsNullOrEmpty(result))
                                            result = result.Replace(
                                                Characters.CarriageReturnString,
                                                String.Empty);

                                        if (!String.IsNullOrEmpty(error))
                                            error = error.Replace(
                                                Characters.CarriageReturnString,
                                                String.Empty);
                                    }

                                    //
                                    // NOTE: Now, "result" contains any StdOut output and "error"
                                    //        contains any StdErr output.
                                    //
                                    if ((code == ReturnCode.Ok) && !background &&
                                        captureOutput && (exitCodeVarName != null))
                                    {
                                        code = interpreter.SetVariableValue(VariableFlags.None,
                                            exitCodeVarName, exitCode.ToString(), null, ref error);
                                    }

                                    if ((code == ReturnCode.Ok) && !useShellExecute &&
                                        !background && captureOutput && (stdOutVarName != null))
                                    {
                                        code = interpreter.SetVariableValue(VariableFlags.None,
                                            stdOutVarName, result, null, ref error);
                                    }

                                    if ((code == ReturnCode.Ok) && !useShellExecute &&
                                        !background && captureOutput && (stdErrVarName != null))
                                    {
                                        code = interpreter.SetVariableValue(VariableFlags.None,
                                            stdErrVarName, error, null, ref error);
                                    }

                                    //
                                    // NOTE: If they specified a "success" exit code, make sure
                                    //       that is the same as the exit code we actually got
                                    //       from the process.
                                    //
                                    if ((code == ReturnCode.Ok) && !background && captureOutput &&
                                        (successExitCode != null) && (exitCode != successExitCode))
                                    {
                                        /* IGNORED */
                                        interpreter.SetVariableValue(
                                            VariableFlags.GlobalOnly, TclVars.ErrorCode,
                                            StringList.MakeList(
                                                "CHILDSTATUS", processId, exitCode),
                                            null);

                                        Engine.SetErrorCodeSet(interpreter, true);

                                        error = "child process exited abnormally";
                                        code = ReturnCode.Error;
                                    }

                                    if (code != ReturnCode.Ok)
                                        //
                                        // NOTE: Transfer error to command result.
                                        //
                                        result = error;
                                }
                                else
                                {
                                    //
                                    // NOTE: Transfer error to command result.
                                    //
                                    result = error;
                                }
                            }
                            else
                            {
                                result = "wrong # args: should be \"exec ?options? arg ?arg ...?\"";
                                code = ReturnCode.Error;
                            }
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"exec ?options? arg ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}
