/*
 * PolicyOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.IO;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Components.Public.Delegates;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Components.Private
{
    [ObjectId("ab00e89a-8a1f-404b-91fd-32d10d0f44ba")]
    internal static class PolicyOps
    {
        #region Private (Internal) Data
        //
        // NOTE: This list defines the policies that are added to every
        //       interpreter.
        //
        internal static ExecuteCallbackDictionary Policies =
            new ExecuteCallbackDictionary(new ExecuteCallback[] {
            FilePolicyCallback, InfoPolicyCallback, InterpPolicyCallback,
            ObjectPolicyCallback, PackagePolicyCallback, SourcePolicyCallback
        });
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Helper Methods
        private static ReturnCode EvaluatePolicyScript(
            Interpreter interpreter,
            string text,
            ArgumentList arguments,
            PolicyFlags policyFlags,
            ref Result result
            ) /* THREAD-SAFE, RE-ENTRANT */
        {
            if (interpreter != null)
            {
                //
                // NOTE: Interpreter is valid, so far, so good.
                //
                ReturnCode code = ReturnCode.Ok;

                //
                // NOTE: Does the caller provide an argument list to
                //       append to the script prior to evaluating it?
                //
                bool appendArguments =
                    FlagOps.HasFlags(policyFlags, PolicyFlags.Arguments, true);

                //
                // NOTE: Does the caller want us to attempt to split
                //       the script into a list?
                //
                if (FlagOps.HasFlags(policyFlags, PolicyFlags.SplitList, true))
                {
                    //
                    // NOTE: We are in "list" mode.  Attempt to parse
                    //       the script as a list and then append the
                    //       supplied argument list before converting
                    //       it back to a string.
                    //
                    StringList list = null;

                    code = Parser.SplitList(interpreter, text, 0,
                        Length.Invalid, false, false, ref list,
                        ref result);

                    if (code == ReturnCode.Ok)
                    {
                        if (appendArguments && (arguments != null))
                            list.Add(arguments);

                        text = list.ToString();
                    }
                }
                else if (appendArguments && (arguments != null))
                {
                    //
                    // NOTE: Arguments were supplied; however, we are
                    //       not operating in "list" mode.  Just append
                    //       the arguments as a single string to the
                    //       script (separated by an intervening space).
                    //
                    text += Characters.Space.ToString() + arguments;
                }

                //
                // NOTE: Did the list parsing code above succeed, if it
                //       was requested?
                //
                if (code == ReturnCode.Ok)
                    code = interpreter.EvaluateScript(text, ref result); /* EXEMPT */

                return code;
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Policy Implementations
        #region Trusted Sub-Command Policy Implementation
        public static ReturnCode SubCommandPolicy( /* POLICY IMPLEMENTATION */
            PolicyFlags flags,
            Type commandType,
            long commandToken,
            StringList subCommands,
            bool allowed,
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            IPolicyContext policyContext = null;
            ICommand command = null;
            bool match = false;

            if (RuntimeOps.ExtractPolicyContextAndCommand(
                    interpreter, clientData, commandType, commandToken,
                    ref policyContext, ref command, ref match,
                    ref result) == ReturnCode.Ok)
            {
                if (arguments != null)
                {
                    if (match)
                    {
                        string subCommand = null;

                        if (arguments.Count >= 2) /* ENSEMBLE */
                            subCommand = arguments[1];

                        if (!String.IsNullOrEmpty(subCommand))
                        {
                            if (RuntimeOps.SubCommandFromEnsemble(command, true,
                                    false, ref subCommand, ref result) == ReturnCode.Ok)
                            {
                                if (allowed)
                                {
                                    if ((subCommands != null) &&
                                        subCommands.Contains(subCommand,
                                            StringOps.SystemStringComparisonType))
                                    {
                                        //
                                        // NOTE: The sub-command is in the
                                        //       "allowed" list, vote to
                                        //       allow the command to be
                                        //       executed.
                                        //
                                        policyContext.Approved();

                                        //
                                        // NOTE: Return the sub-command as
                                        //       the result.
                                        //
                                        policyContext.Result = subCommand;
                                    }
                                }
                                else
                                {
                                    if ((subCommands != null) &&
                                        !subCommands.Contains(subCommand,
                                            StringOps.SystemStringComparisonType))
                                    {
                                        //
                                        // NOTE: The sub-command is not in
                                        //       the "denied" list, vote to
                                        //       allow the command to be
                                        //       executed.
                                        //
                                        policyContext.Approved();

                                        //
                                        // NOTE: Return the sub-command as
                                        //       the result.
                                        //
                                        policyContext.Result = subCommand;
                                    }
                                }
                            }
                        }
                    }

                    //
                    // NOTE: The policy checking has been successful;
                    //       however, this does not necessarily mean
                    //       that we allow the command to be executed.
                    //
                    return ReturnCode.Ok;
                }
                else
                {
                    result = "invalid argument list";
                }
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Trusted URI Policy Implementation
        public static ReturnCode UriPolicy( /* POLICY IMPLEMENTATION */
            PolicyFlags flags,
            Type commandType,
            long commandToken,
            Uri uri,
            UriDictionary<object> uris,
            bool allowed,
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            IPolicyContext policyContext = null;
            bool match = false;

            if (RuntimeOps.ExtractPolicyContextAndCommand(
                    interpreter, clientData, commandType,
                    commandToken, ref policyContext, ref match,
                    ref result) == ReturnCode.Ok)
            {
                if (match)
                {
                    if (allowed)
                    {
                        if ((uris != null) &&
                            uris.ContainsSchemeAndServer(uri))
                        {
                            //
                            // NOTE: The URI is in the "allowed"
                            //       list, vote to allow the
                            //       command to be executed.
                            //
                            policyContext.Approved();
                        }
                    }
                    else
                    {
                        if ((uris != null) &&
                            !uris.ContainsSchemeAndServer(uri))
                        {
                            //
                            // NOTE: The URI is not in the
                            //       "denied" list, vote to
                            //       allow the command to be
                            //       executed.
                            //
                            policyContext.Approved();
                        }
                    }
                }

                //
                // NOTE: The policy checking has been successful;
                //       however, this does not necessarily mean
                //       that we allow the command to be executed.
                //
                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Trusted Directory Policy Implementation
        public static ReturnCode DirectoryPolicy( /* POLICY IMPLEMENTATION */
            PolicyFlags flags,
            Type commandType,
            long commandToken,
            string fileName,
            PathDictionary<object> directories,
            bool allowed,
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            IPolicyContext policyContext = null;
            bool match = false;

            if (RuntimeOps.ExtractPolicyContextAndCommand(
                    interpreter, clientData, commandType,
                    commandToken, ref policyContext, ref match,
                    ref result) == ReturnCode.Ok)
            {
                if (match)
                {
                    string directory = null;

                    try
                    {
                        /* throw */
                        directory = Path.GetDirectoryName(
                            PathOps.BaseDirectorySubstitution(fileName));
                    }
                    catch
                    {
                        // do nothing.
                    }

                    if (!String.IsNullOrEmpty(directory))
                    {
                        if (allowed)
                        {
                            if ((directories != null) &&
                                directories.Contains(directory))
                            {
                                //
                                // NOTE: The directory is in the
                                //       "allowed" list, vote to
                                //       allow the command to be
                                //       executed.
                                //
                                policyContext.Approved();
                            }
                        }
                        else
                        {
                            if ((directories != null) &&
                                !directories.Contains(directory))
                            {
                                //
                                // NOTE: The directory is not in
                                //       the "denied" list, vote
                                //       to allow the command to
                                //       be executed.
                                //
                                policyContext.Approved();
                            }
                        }
                    }
                }

                //
                // NOTE: The policy checking has been successful;
                //       however, this does not necessarily mean
                //       that we allow the command to be executed.
                //
                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Trusted Object Type Policy Implementation
        public static ReturnCode TypePolicy( /* POLICY IMPLEMENTATION */
            PolicyFlags flags,
            Type commandType,
            long commandToken,
            Type objectType,
            TypeList types,
            bool allowed,
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            IPolicyContext policyContext = null;
            bool match = false;

            if (RuntimeOps.ExtractPolicyContextAndCommand(
                    interpreter, clientData, commandType,
                    commandToken, ref policyContext, ref match,
                    ref result) == ReturnCode.Ok)
            {
                if (match)
                {
                    if (objectType != null)
                    {
                        if (allowed)
                        {
                            if ((types != null) &&
                                types.Contains(objectType))
                            {
                                //
                                // NOTE: The type is in the
                                //       "allowed" list, vote to
                                //       allow the command to be
                                //       executed.
                                //
                                policyContext.Approved();
                            }
                        }
                        else
                        {
                            if ((types != null) &&
                                !types.Contains(objectType))
                            {
                                //
                                // NOTE: The type is not in the
                                //       "denied" list, vote to
                                //       allow the command to be
                                //       executed.
                                //
                                policyContext.Approved();
                            }
                        }
                    }
                }

                //
                // NOTE: The policy checking has been successful;
                //       however, this does not necessarily mean
                //       that we allow the command to be executed.
                //
                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Dynamic User Managed Callback Policy Implementation
        public static ReturnCode CallbackPolicy( /* POLICY IMPLEMENTATION */
            PolicyFlags flags,
            Type commandType,
            long commandToken,
            ICallback callback,
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            IPolicyContext policyContext = null;
            bool match = false;

            if (RuntimeOps.ExtractPolicyContextAndCommand(
                    interpreter, clientData, commandType,
                    commandToken, ref policyContext, ref match,
                    ref result) == ReturnCode.Ok)
            {
                if (arguments != null)
                {
                    if (match)
                    {
                        if (callback != null)
                        {
                            ReturnCode localCode;
                            Result localResult = null;

                            localCode = callback.Invoke(
                                new StringList(arguments), ref localResult);

                            switch (localCode)
                            {
                                case ReturnCode.Ok:
                                    {
                                        //
                                        // NOTE: The callback was executed
                                        //       successfully, vote to allow
                                        //       the command to be executed.
                                        //
                                        policyContext.Approved();
                                        break;
                                    }
                                case ReturnCode.Break:
                                    {
                                        //
                                        // NOTE: No vote is made for this
                                        //       return code.
                                        //
                                        break;
                                    }
                                case ReturnCode.Continue:
                                    {
                                        //
                                        // NOTE: This return code represents
                                        //       an official "undecided" vote.
                                        //
                                        policyContext.Undecided();
                                        break;
                                    }
                                case ReturnCode.Error:
                                case ReturnCode.Return: // NOTE: Invalid policy return code.
                                default:
                                    {
                                        //
                                        // NOTE: An error or any other
                                        //       return code is interpreted
                                        //       as a "denied" vote.
                                        //
                                        policyContext.Denied();
                                        break;
                                    }
                            }

                            //
                            // NOTE: Set the informational policy result to
                            //       the result of the callback execution.
                            //
                            policyContext.Result = Result.Copy(
                                localCode, localResult, true); /* COPY */
                        }
                    }

                    //
                    // NOTE: The policy checking has been successful;
                    //       however, this does not necessarily mean
                    //       that we allow the command to be executed.
                    //
                    return ReturnCode.Ok;
                }
                else
                {
                    result = "invalid argument list";
                }
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Dynamic User Script Evaluation Policy Implementation
        public static ReturnCode ScriptPolicy( /* POLICY IMPLEMENTATION */
            PolicyFlags flags,
            Type commandType,
            long commandToken,
            Interpreter policyInterpreter,
            string text,
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            IPolicyContext policyContext = null;
            bool match = false;

            if (RuntimeOps.ExtractPolicyContextAndCommand(
                    interpreter, clientData, commandType,
                    commandToken, ref policyContext, ref match,
                    ref result) == ReturnCode.Ok)
            {
                if (match)
                {
                    //
                    // NOTE: *WARNING* Empty scripts are allowed, please
                    //        do not change this to "!String.IsNullOrEmpty".
                    //
                    if ((policyInterpreter != null) && (text != null))
                    {
                        ReturnCode localCode;
                        Result localResult = null;

                        localCode = EvaluatePolicyScript(
                            policyInterpreter, text, arguments, flags, ref localResult);

                        switch (localCode)
                        {
                            case ReturnCode.Ok:
                                {
                                    //
                                    // NOTE: The callback was executed
                                    //       successfully, vote to allow
                                    //       the command to be executed.
                                    //
                                    policyContext.Approved();
                                    break;
                                }
                            case ReturnCode.Break:
                                {
                                    //
                                    // NOTE: No vote is made for this
                                    //       return code.
                                    //
                                    break;
                                }
                            case ReturnCode.Continue:
                                {
                                    //
                                    // NOTE: This return code represents
                                    //       an official "undecided" vote.
                                    //
                                    policyContext.Undecided();
                                    break;
                                }
                            case ReturnCode.Error:
                            case ReturnCode.Return: // NOTE: Invalid policy return code.
                            default:
                                {
                                    //
                                    // NOTE: An error or any other
                                    //       return code is interpreted
                                    //       as a "denied" vote.
                                    //
                                    policyContext.Denied();
                                    break;
                                }
                        }

                        //
                        // NOTE: Set the informational policy result to
                        //       the result of the script evaluation.
                        //
                        policyContext.Result = Result.Copy(
                            localCode, localResult, true); /* COPY */
                    }
                }

                //
                // NOTE: The policy checking has been successful;
                //       however, this does not necessarily mean
                //       that we allow the command to be executed.
                //
                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Default Core Command Policies
        #region The Default [file] Command Policy
        [MethodFlags(MethodFlags.CommandPolicy | MethodFlags.NoAdd)]
        private static ReturnCode FilePolicyCallback( /* POLICY */
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            StringList subCommands = new StringList(new string[] {
                "dirname", "join", "split"
            });

            return SubCommandPolicy(
                PolicyFlags.SubCommand, typeof(_Commands._File), 0,
                subCommands, true, interpreter, clientData, arguments,
                ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region The Default [info] Command Policy
        [MethodFlags(MethodFlags.CommandPolicy | MethodFlags.NoAdd)]
        private static ReturnCode InfoPolicyCallback( /* POLICY */
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            StringList subCommands = new StringList(new string[] {
                "commands", "engine", "exists", "globals", "level",
                "library", "procs", "script", "vars"
            });

            return SubCommandPolicy(
                PolicyFlags.SubCommand, typeof(_Commands.Info), 0,
                subCommands, true, interpreter, clientData, arguments,
                ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region The Default [interp] Command Policy
        [MethodFlags(MethodFlags.CommandPolicy | MethodFlags.NoAdd)]
        private static ReturnCode InterpPolicyCallback( /* POLICY */
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            StringList subCommands = new StringList(new string[] {
                "cancel", "exists", "issafe"
            });

            return SubCommandPolicy(
                PolicyFlags.SubCommand, typeof(_Commands.Interp), 0,
                subCommands, true, interpreter, clientData, arguments,
                ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region The Default [object] Command Policy
        [MethodFlags(MethodFlags.CommandPolicy | MethodFlags.NoAdd)]
        private static ReturnCode ObjectPolicyCallback( /* POLICY */
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            StringList subCommands = new StringList(new string[] {
                "invoke", "invokeall"
            });

            return SubCommandPolicy(
                PolicyFlags.SubCommand, typeof(_Commands.Object), 0,
                subCommands, true, interpreter, clientData, arguments,
                ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region The Default [package] Command Policy
        [MethodFlags(MethodFlags.CommandPolicy | MethodFlags.NoAdd)]
        private static ReturnCode PackagePolicyCallback( /* POLICY */
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            StringList subCommands = new StringList(new string[] {
                "indexes", "reset", "scan"
            });

            return SubCommandPolicy(
                PolicyFlags.SubCommand, typeof(_Commands.Package), 0,
                subCommands, false, interpreter, clientData, arguments,
                ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region The Default [source] Command Policy
        [MethodFlags(MethodFlags.CommandPolicy | MethodFlags.NoAdd)]
        private static ReturnCode SourcePolicyCallback( /* POLICY */
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            //
            // HACK: There is a small problem with this policy.  We need to
            //       examine the first argument to determine which policy
            //       checking method we would like to use; however, we do
            //       not actually know if the command being executed is
            //       [source].  In practice, this does not matter because
            //       the policy methods themselves double-check the command
            //       type.  Therefore, the argument checking here will be
            //       pointless (but harmless) if the command is not [source].
            //
            string fileName = null;

            if ((arguments != null) && (arguments.Count >= 2))
                fileName = arguments[arguments.Count - 1];

            //
            // NOTE: If the file name represents a remote URI, use slightly
            //       different policy handling.
            //
            Uri uri = null;

            if (PathOps.IsRemoteUri(fileName, ref uri))
            {
                //
                // NOTE: Only allow remote sites that we know, trust, and have
                //       100% positive control over.
                //
                UriDictionary<object> trustedUris = new UriDictionary<object>();

                RuntimeOps.AddTrustedUris(interpreter, trustedUris);

                return UriPolicy(
                    PolicyFlags.Uri, typeof(_Commands.Source), 0, uri,
                    trustedUris, true, interpreter, clientData, arguments,
                    ref result);
            }
            else
            {
                PathDictionary<object> directories = new PathDictionary<object>();

                RuntimeOps.AddTrustedDirectories(interpreter, directories);

                return DirectoryPolicy(
                    PolicyFlags.Directory, typeof(_Commands.Source), 0,
                    fileName, directories, true, interpreter, clientData,
                    arguments, ref result);
            }
        }
        #endregion
        #endregion
    }
}
